/*-
 * Copyright (C)2022 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2022 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: style_copy.c,v 1.21 2022/03/26 22:17:10 babolo Exp $"

#define BLIN_COMPAT      4
#define Bpars_COMPAT     3
#define MULAR_COMPAT     0
#define MIFE_COMPAT      4
#define PGOBLIN_COMPAT   4
#define PGOBLIN_INTERNAL 1
#define PGOBLIN_MODULE   pgoblin_module

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "pgoblin.h"

#define PGOBLIN_STYLE_ID "#pGoblin-copy"

struct pgoblin_realsty {
    char              id[PGOBLIN_STRING_ID_LEN];  /*  "#pGoblin"             */
    u_int32_t         flags;
#   define            PGOBLIN_COPE  0x000100
};

static int
/**********************************************************************
 **                                                                  **/
parser(pgoblin_exenv *exenv, pgoblin_nr nst, char *srcstyle) {      /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
#   define         ST     (rst->style)
    pgoblin_rst   *rst;
    int            ex = EX_OK;     /*                                                     */

    GET_STY(rst, exenv->options, nst);
    if  (!ST) {
        if  (!(rst->style = calloc(1, sizeof(pgoblin_realsty)))) {
            ifBLIN_QX0("No memory #1");
            ERROUT(EX_OSERR, ENOMEM);
        }
        strncpy(ST->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID));
        ST->flags = pgoblin_default;
    } else if (!!strncmp(ST->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
            ifBLIN_QX0("Illegal style %s", ST->id);
            ERROUT(EX_DATAERR, EINVAL);
    }
    ST->flags &= ~BLIN_MASK;
    for (int i = 0; !!srcstyle[i]; ++i) {
        if  (' ' == srcstyle[i]) continue;
        if  ('E' == srcstyle[i]) ST->flags |= PGOBLIN_COPE;
        break;
    }
out:
    return(ex);
#   undef  ST
#   undef blin_internal_flags
}

static void
/**********************************************************************
 **                                                                  **/
dump(pgoblin_exenv *exenv, pgoblin_nr nst) {                        /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    return;
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
clear(pgoblin_exenv *exenv, pgoblin_nr nst) {                       /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    pgoblin_rst     *rst;
    int              ex = EX_OK;
    pgoblin_realsty *st;

    GET_STY(rst, exenv->options, nst);
    if  ((st = rst->style)) {
        if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
            ifBLIN_QX0("Illegal style %s", st->id);
            ERROUT(EX_DATAERR, EINVAL);
        }
        free(st);
        st = NULL;
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static char *
/**********************************************************************
 **                                                                  **/
get(pgoblin_exenv *exenv, pgoblin_nr nst, pgoblin_nr nct) {         /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    char       *found = NULL;

    return(found);
#   undef blin_internal_flags
}

static ssize_t
/**********************************************************************
 **                                                                  **/
table( pgoblin_exenv *exenv                                         /**/
     , pgoblin_nr     nst                                           /**/
     , pgoblin_nr     nou                                           /**/
     , pgoblin_nr     nin                                           /**/
     , ssize_t        low                                           /**/
     , ssize_t        hig                                           /**/
     ) {                                                            /**
 **                                                                  **
 **********************************************************************/
#   define blin_internal_flags (exenv->options->flags & BLIN_MASK)
    ssize_t          textlength = 0; /*                                      */
    ssize_t          fullrownum = 0; /*                                    */
    ssize_t          hirowborder;    /* ,                */
    ssize_t          tbwidth;        /*                                                 */
    pgoblin_rio     *rou;
    pgoblin_rio *rin;
    pgoblin_rst     *rst;
    ssize_t          ex = 0;
    pgoblin_realsty *st;

    GET_STY(rst, exenv->options, nst);
    GET_RIO(rou, exenv->options, nou);
    GET_RIO(rin, exenv->options, nin);
    st = rst->style;
    if  (!!strncmp(st->id, PGOBLIN_STYLE_ID, strlen(PGOBLIN_STYLE_ID))) {
        ifBLIN_QX0("Illegal style %s", st->id);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (low >= 0) fullrownum = low;
    hirowborder = fullrownum + pgoblin_db_resinfo(exenv, nin, PGOBLIN_Ntuples);
    tbwidth = pgoblin_db_resinfo(exenv, nin, PGOBLIN_Nfields);
    for (ssize_t lorow = fullrownum; fullrownum < hirowborder; fullrownum++) {
        for (int colnum = 0; colnum < tbwidth; colnum++) {
            ssize_t localrownum = fullrownum - lorow;
            char   *buf;
            char   *c;
            ssize_t l;
            ssize_t k;

            if  (ex < 0) goto out;
            if  (!pgoblin_db_valinfo(exenv, nin, PGOBLIN_IsNull, localrownum, colnum)) {
                l = pgoblin_db_valinfo(exenv, nin, PGOBLIN_Length, localrownum, colnum);
                c = pgoblin_db_getvalue(exenv, nin, localrownum, colnum);
            } else {
                l = 0;
                c = NULL;
            }
            if  (!(buf = malloc((c ? l * 2 : 2) + 2))) {
                ifBLIN_QW0("No mem");
                ex = -EX_OSERR;
                goto out;
            }
            if  (0 > (k = pgoblin_onecopy(buf, c, l, (colnum < (tbwidth - 1)) ? -1 : 0))) {
                ex = k;
                goto out;
            }
            buf[k] = 0;
            ex = pgoblin_wr(rou, buf, k);
            free(buf);
            if  (ex < 0) goto out;
            textlength += ex;
    }   }
    if  (st->flags & PGOBLIN_COPE) ex = pgoblin_wr(rou, "\\.\n", 3);
out:
    return((ex < 0) ? ex : textlength);
#   undef blin_internal_flags
}

static int
/**********************************************************************
 **                                                                  **/
init(pgoblin_main *options, int slot) {                             /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

static int
/**********************************************************************
 **                                                                  **/
fini(pgoblin_main *options) {                                       /**
 **                                                                  **
 **********************************************************************/
    return(0);
}

pgoblin_styles PGOBLIN_MODULE =
{ 0
, "5pGoblin-" VERS
, "copy\0" VERS
, 0
, init
, fini
, parser
, dump
, clear
, get
, table
, NULL /* load   */
, NULL /* store  */
, NULL /* dummy0 */
, NULL /* dummy1 */
};
